package net.kierenb.mapcast;

import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashSet;
import java.util.LinkedHashSet;
import java.util.Map;
import java.util.Set;

import net.kierenb.mapcast.replication.EntryDao;
import net.kierenb.mapcast.serializer.KeySerializer;
import net.kierenb.mapcast.serializer.ValueSerializer;
import net.kierenb.mapcast.sql.ResultTransformer;
import net.kierenb.mapcast.sql.SqlUtil;
import net.kierenb.mapcast.sql.StatementPreparator;
import net.kierenb.mapcast.util.TimeBasedId;

/**
 * Copyright 2012 Kieren Beckett
 * 
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * 
 * http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
 * an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
 * specific language governing permissions and limitations under the License.
 * 
 * @author mail@kierenb.net
 */
class ReplicatedMapImpl<K, V> implements ReplicatedMap<K, V> {

    private TimeBasedId timeId = new TimeBasedId();

    private String name;
    private String node;
    private SqlUtil sqlUtil;
    private KeySerializer<K> keySerializer;
    private ValueSerializer<V> valueSerializer;
    private Map<String, Long> lagMonitoringMap;

    ReplicatedMapImpl(String name, String node, SqlUtil sqlUtil, KeySerializer<K> keySerializer, ValueSerializer<V> valueSerializer, Map<String, Long> lagMonitoringMap) {
        if (name == null) {
            throw new IllegalArgumentException("Map name cannot be null");
        }

        if (name.length() > EntryDao.MAX_MAP_NAME_LENGTH) {
            throw new IllegalArgumentException("Map name length too long");
        }

        if (keySerializer == null) {
            throw new IllegalArgumentException("Key serializer cannot be null");
        }

        if (valueSerializer == null) {
            throw new IllegalArgumentException("Value serializer cannot be null");
        }

        this.name = name;
        this.node = node;
        this.sqlUtil = sqlUtil;
        this.keySerializer = keySerializer;
        this.valueSerializer = valueSerializer;
        this.lagMonitoringMap = lagMonitoringMap;
    }

    @Override
    public V get(final Object key) {
        byte[] bytes = getRaw(key);
        if (bytes == null) {
            return null;
        }
        return valueSerializer.toObject(bytes);
    }

    @Override
    public byte[] getRaw(final Object key) {
        if (keySerializer.fromObject(key) == null) {
            return null;
        }

        return sqlUtil.executeQuery("SELECT key, value FROM entry WHERE map=? AND key=? AND deletedTime=-1", new StatementPreparator() {
            @Override
            public void prepare(PreparedStatement stmt) throws SQLException {
                int i = 1;
                stmt.setString(i++, name);
                stmt.setString(i++, keySerializer.fromObject(key));
            }
        }, new ResultTransformer<byte[]>() {
            @Override
            public byte[] transform(ResultSet rs) throws SQLException {
                if (rs.next()) {
                    return rs.getBytes("value");
                }
                return null;
            }
        });
    }
    
    @Override
    public V put(final K key, final V value) {
        putRaw(key, valueSerializer.fromObject(value));
        return value;
    }

    @Override
    public void putRaw(final K key, final byte[] bytes) {
        sqlUtil.executeUpdate("MERGE INTO entry USING (VALUES(?, ?, CAST(? AS BLOB), CAST(-1 AS BIGINT), ?, CAST(? AS BIGINT))) AS new(map,key,value,deletedTime,lastUpdatedNode,lastUpdatedTime) ON entry.map = new.map AND entry.key = new.key WHEN MATCHED THEN UPDATE SET entry.value = new.value, entry.deletedTime=new.deletedTime, entry.lastUpdatedNode = new.lastUpdatedNode, entry.lastUpdatedTime = new.lastUpdatedTime WHEN NOT MATCHED THEN INSERT VALUES new.map, new.key, new.value, new.deletedTime, new.lastUpdatedNode, new.lastUpdatedTime", new StatementPreparator() {
            @Override
            public void prepare(PreparedStatement stmt) throws SQLException {
                int i = 1;
                stmt.setString(i++, name);
                stmt.setString(i++, keySerializer.fromObject(key));
                stmt.setBytes(i++, bytes);
                stmt.setString(i++, node);
                stmt.setLong(i++, timeId.getId());
            }
        });

        if (lagMonitoringMap != null) {
            lagMonitoringMap.put(node, System.currentTimeMillis());
        }
    }

    @Override
    public void clear() {
        sqlUtil.executeUpdate("UPDATE entry SET deletedTime=?, lastUpdatedNode=?, lastUpdatedTime=? WHERE map = ?", new StatementPreparator() {
            @Override
            public void prepare(PreparedStatement stmt) throws SQLException {
                int i = 1;
                stmt.setLong(i++, System.currentTimeMillis());
                stmt.setString(i++, node);
                stmt.setLong(i++, timeId.getId());
                stmt.setString(i++, name);
            }
        });

        if (lagMonitoringMap != null) {
            lagMonitoringMap.put(node, System.currentTimeMillis());
        }
    }

    @Override
    public boolean containsKey(final Object key) {
        if (keySerializer.fromObject(key) == null) {
            return false;
        }

        return sqlUtil.executeQuery("SELECT key FROM entry WHERE map=? AND key=? AND deletedTime=-1", new StatementPreparator() {
            @Override
            public void prepare(PreparedStatement stmt) throws SQLException {
                int i = 1;
                stmt.setString(i++, name);
                stmt.setString(i++, keySerializer.fromObject(key));
            }
        }, ResultTransformer.HAS_ROWS);
    }

    @Override
    public boolean containsValue(Object value) {
        for (V comparing : values()) {
            if (comparing == null && value == null) {
                return true;
            }
            if (comparing.equals(value)) {
                return true;
            }
        }
        return false;
    }

    @Override
    public Set<Map.Entry<K, V>> entrySet() {
        return sqlUtil.executeQuery("SELECT key, value FROM entry WHERE deletedTime=-1 AND map=?", new StatementPreparator() {
            @Override
            public void prepare(PreparedStatement stmt) throws SQLException {
                int i = 1;
                stmt.setString(i++, name);
            }
        }, new ResultTransformer<Set<Map.Entry<K, V>>>() {
            @Override
            public Set<Map.Entry<K, V>> transform(ResultSet rs) throws SQLException {
                Set<Map.Entry<K, V>> entrySet = new HashSet<Map.Entry<K, V>>();
                while (rs.next()) {
                    K key = keySerializer.toObject(rs.getString("key"));
                    V value = valueSerializer.toObject(rs.getBytes("value"));
                    entrySet.add(new Entry<K, V>(key, value));
                }
                return entrySet;
            }
        });
    }

    @Override
    public boolean isEmpty() {
        return !sqlUtil.executeQuery("SELECT key FROM entry WHERE deletedTime=-1 AND map=? LIMIT 1", new StatementPreparator() {
            @Override
            public void prepare(PreparedStatement stmt) throws SQLException {
                int i = 1;
                stmt.setString(i++, name);
            }
        }, ResultTransformer.HAS_ROWS);
    }

    @Override
    public Set<K> keySet() {
        return sqlUtil.executeQuery("SELECT key FROM entry WHERE deletedTime=-1 AND map=? ORDER BY key ASC", new StatementPreparator() {
            @Override
            public void prepare(PreparedStatement stmt) throws SQLException {
                int i = 1;
                stmt.setString(i++, name);
            }
        }, KEY_SET_TRANSFORMER);
    }

    @Override
    public Set<K> keySet(final Integer limit) {
        return sqlUtil.executeQuery("SELECT key FROM entry WHERE deletedTime=-1 AND map=? ORDER BY key ASC LIMIT ?", new StatementPreparator() {
            @Override
            public void prepare(PreparedStatement stmt) throws SQLException {
                int i = 1;
                stmt.setString(i++, name);
                stmt.setInt(i++, limit);
            }
        }, KEY_SET_TRANSFORMER);
    }

    @Override
    public Set<K> keySet(final Object keyFromExclusive, final Integer limit) {
        return sqlUtil.executeQuery("SELECT key FROM entry WHERE deletedTime=-1 AND map=? AND key > ? ORDER BY key ASC LIMIT ?", new StatementPreparator() {
            @Override
            public void prepare(PreparedStatement stmt) throws SQLException {
                int i = 1;
                stmt.setString(i++, name);
                stmt.setString(i++, keySerializer.fromObject(keyFromExclusive));
                stmt.setInt(i++, limit);
            }
        }, KEY_SET_TRANSFORMER);
    }

    @Override
    public void putAll(Map<? extends K, ? extends V> m) {
        for (Map.Entry<? extends K, ? extends V> entry : m.entrySet()) {
            put(entry.getKey(), entry.getValue());
        }
    }

    @Override
    public V remove(final Object key) {
        if (keySerializer.fromObject(key) == null) {
            return null;
        }

        V value = get(key);
        sqlUtil.executeUpdate("UPDATE entry SET deletedTime=?, lastUpdatedNode=?, lastUpdatedTime=? WHERE map=? AND key=?", new StatementPreparator() {
            @Override
            public void prepare(PreparedStatement stmt) throws SQLException {
                int i = 1;
                stmt.setLong(i++, System.currentTimeMillis());
                stmt.setString(i++, node);
                stmt.setLong(i++, timeId.getId());
                stmt.setString(i++, name);
                stmt.setString(i++, keySerializer.fromObject(key));
            }
        });

        if (lagMonitoringMap != null) {
            lagMonitoringMap.put(node, System.currentTimeMillis());
        }

        return value;
    }

    @Override
    public int size() {
        return sqlUtil.executeQuery("SELECT count(*) FROM entry WHERE deletedTime=-1 AND map=?", new StatementPreparator() {
            @Override
            public void prepare(PreparedStatement stmt) throws SQLException {
                int i = 1;
                stmt.setString(i++, name);
            }
        }, new ResultTransformer<Integer>() {
            @Override
            public Integer transform(ResultSet rs) throws SQLException {
                rs.next();
                return rs.getInt(1);
            }
        });
    }

    @Override
    public Collection<V> values() {
        return sqlUtil.executeQuery("SELECT value FROM entry WHERE deletedTime=-1 AND map=?", new StatementPreparator() {
            @Override
            public void prepare(PreparedStatement stmt) throws SQLException {
                int i = 1;
                stmt.setString(i++, name);
            }
        }, new ResultTransformer<Collection<V>>() {
            @Override
            public Collection<V> transform(ResultSet rs) throws SQLException {
                Collection<V> values = new ArrayList<V>();
                while (rs.next()) {
                    V value = valueSerializer.toObject(rs.getBytes("value"));
                    values.add(value);
                }
                return values;
            }
        });
    }

    private final ResultTransformer<Set<K>> KEY_SET_TRANSFORMER = new ResultTransformer<Set<K>>() {
        @Override
        public Set<K> transform(ResultSet rs) throws SQLException {
            Set<K> keys = new LinkedHashSet<K>();
            while (rs.next()) {
                K key = keySerializer.toObject(rs.getString("key"));
                keys.add(key);
            }
            return keys;
        }
    };

    private static final class Entry<J, W> implements Map.Entry<J, W> {
        private J key;
        private W value;

        public Entry(J key, W value) {
            this.key = key;
            this.value = value;
        }

        @Override
        public J getKey() {
            return key;
        }

        @Override
        public W getValue() {
            return value;
        }

        @Override
        public W setValue(W value) {
            this.value = value;
            return value;
        }

        @Override
        public int hashCode() {
            final int prime = 31;
            int result = 1;
            result = prime * result + ((key == null) ? 0 : key.hashCode());
            return result;
        }

        @Override
        public boolean equals(Object obj) {
            if (this == obj) return true;
            if (obj == null) return false;
            if (getClass() != obj.getClass()) return false;
            Entry<?, ?> other = (Entry<?, ?>)obj;
            if (key == null) {
                if (other.key != null) return false;
            }
            else if (!key.equals(other.key)) return false;
            return true;
        }
    }

}
