package net.kierenb.mapcast;

import java.io.File;
import java.io.Serializable;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Date;
import java.util.HashMap;
import java.util.Map;
import java.util.Set;

import net.kierenb.mapcast.api.LagResource;
import net.kierenb.mapcast.api.MapResource;
import net.kierenb.mapcast.discovery.NodeUrlDao;
import net.kierenb.mapcast.discovery.NodeUrlDiscovery;
import net.kierenb.mapcast.replication.EntryDao;
import net.kierenb.mapcast.replication.EntryPoller;
import net.kierenb.mapcast.replication.EntryResource;
import net.kierenb.mapcast.replication.StateDao;
import net.kierenb.mapcast.serializer.KeySerializer;
import net.kierenb.mapcast.serializer.StringKeySerializer;
import net.kierenb.mapcast.serializer.ValueSerializer;
import net.kierenb.mapcast.serializer.XmlValueSerializer;
import net.kierenb.mapcast.sql.SqlUtil;
import net.kierenb.mapcast.util.ConfigDao;
import net.kierenb.mapcast.util.HttpResource;
import net.kierenb.mapcast.util.HttpServer;
import net.kierenb.mapcast.util.PeriodicTask;

import org.apache.log4j.Logger;

/**
 * A replicated map server. Generally a JVM will have one map server hosting one or more maps. There can however be
 * multiple servers in a single JVM if there's a need to further logically separate groups of maps. Having created a map
 * server, maps can be created/retrieved from it by calling getMap().
 * 
 * Copyright 2012 Kieren Beckett
 * 
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * 
 * http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
 * an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
 * specific language governing permissions and limitations under the License.
 * 
 * @author mail@kierenb.net
 */
public class ReplicatedMapServer {
    private static final Logger log = Logger.getLogger(ReplicatedMapServer.class);
    
    public static final String LAG_MAP_NAME = "lagMonitoring";
    public static final Integer MAX_CONTEXT_LENGTH = 40;

    private Integer port = 0;
    private Integer entryPollingPeriod = 30 * 1000;
    private Integer taskPollingPeriod = 60 * 1000;
    private Integer socketTimeout = 10 * 1000;
    private Integer trimDeletedEntriesOlderThan = 7 * 24 * 60 * 60 * 1000;
    private Integer trimUnadvertisedUrlsOlderThan = 5 * 60 * 1000;

    private String context;
    private SqlUtil sqlUtil;
    private ConfigDao configDao;
    private EntryDao entryDao;
    private StateDao stateDao;
    private NodeUrlDao nodeUrlDao;
    private String node;
    private Map<String, Long> lagMonitoringMap;

    private HttpServer httpServer;
    private EntryPoller entryPoller;
    private NodeUrlDiscovery nodeUrlDiscovery;
    private PeriodicTask periodicTasks;

    /**
     * Create a new replicated map server. Don't forget to start/stop it.
     * 
     * @param context A name that uniquely identifies a cluster. This should be set to the same value across a number of
     *            servers for them to replicate with each other. Maximum 40 characters.
     */
    public ReplicatedMapServer(String context) {
        this(context, context + File.separator + "db");
    }

    /**
     * Create a new replicated map server with a custom database location. Don't forget to start/stop it.
     * 
     * @param context A name that uniquely identifies a cluster. This should be set to the same value across a number of
     *            servers for them to replicate with each other. Maximum 40 characters.
     * @param dbFilename The filename used for this servers local database. A number of files with different suffixes
     *            appended to this filename will be created.
     */
    public ReplicatedMapServer(String context, String dbFilename) {
        if (context == null) {
            throw new IllegalArgumentException("Context cannot be null");
        }

        if (context.length() > MAX_CONTEXT_LENGTH) {
            throw new IllegalArgumentException("Context length too long");
        }

        if (dbFilename == null) {
            throw new IllegalArgumentException("Database filename cannot be null");
        }

        this.context = context;
        this.sqlUtil = new SqlUtil(dbFilename);
        this.configDao = new ConfigDao(sqlUtil);
        this.entryDao = new EntryDao(sqlUtil);
        this.stateDao = new StateDao(sqlUtil);
        this.nodeUrlDao = new NodeUrlDao(sqlUtil);
        this.node = configDao.getNode();
        this.lagMonitoringMap = new ReplicatedMapImpl<String, Long>(LAG_MAP_NAME, node, sqlUtil, new StringKeySerializer(), new XmlValueSerializer<Long>(), null);
    }

    /**
     * Get the context that was provided when creating this server.
     * 
     * @return The context that was provided when creating this server.
     */
    public String getContext() {
        return context;
    }

    /**
     * Get a replicated map. If the map name is new the map will be created.
     * 
     * @param <V> The type of value stored in this map.
     * @param name A name that uniquely identifies this map and is used to reference it across the cluster. Maximum 20
     *            characters.
     * @return A replicated map.
     */
    public <V extends Serializable> ReplicatedMap<String, V> getMap(String name) {
        return getMap(name, new StringKeySerializer(), new XmlValueSerializer<V>());
    }

    /**
     * Get a replicated map with a custom key serializer. If the map name is new the map will be created.
     * 
     * @param <K> The type of key stored in this map.
     * @param <V> The type of value stored in this map.
     * @param name A name that uniquely identifies this map and is used to reference it across the cluster. Maximum 20
     *            characters.
     * @param keySerializer The serializer used to serialize keys in this map. This must be the same for the same map on
     *            other nodes.
     * @return A replicated map.
     */
    public <K, V extends Serializable> ReplicatedMap<K, V> getMap(String name, KeySerializer<K> keySerializer) {
        return getMap(name, keySerializer, new XmlValueSerializer<V>());
    }

    /**
     * Get a replicated map with a custom value serializer. If the map name is new the map will be created.
     * 
     * @param <V> The type of value stored in this map.
     * @param name A name that uniquely identifies this map and is used to reference it across the cluster. Maximum 20
     *            characters.
     * @param valueSerializer The serializer used to serialize values in this map. This must be the same for the same
     *            map on other nodes.
     * @return A replicated map.
     */
    public <V extends Serializable> ReplicatedMap<String, V> getMap(String name, ValueSerializer<V> valueSerializer) {
        return getMap(name, new StringKeySerializer(), valueSerializer);
    }

    /**
     * Get a replicated map with a custom key and value serializer. If the map name is new the map will be created.
     * 
     * @param <K> The type of key stored in this map.
     * @param <V> The type of value stored in this map.
     * @param name A name that uniquely identifies this map and is used to reference it across the cluster. Maximum 20
     *            characters.
     * @param keySerializer The serializer used to serialize keys in this map. This must be the same for the same map on
     *            other nodes.
     * @param valueSerializer The serializer used to serialize values in this map. This must be the same for the same
     *            map on other nodes.
     * @return A replicated map.
     */
    public <K, V> ReplicatedMap<K, V> getMap(String name, KeySerializer<K> keySerializer, ValueSerializer<V> valueSerializer) {
        if (ReplicatedMapServer.LAG_MAP_NAME.equals(name)) {
            throw new IllegalArgumentException("Map name is a reserved word");
        }

        return new ReplicatedMapImpl<K, V>(name, node, sqlUtil, keySerializer, valueSerializer, lagMonitoringMap);
    }

    /**
     * Get how behind real time replication from remote nodes is for monitoring purposes. If all is well the lag should
     * be no more than a minute or two.
     * 
     * @return A map of remote node names to the time of their last replication sync with this node.
     */
    public Map<String, Long> getLag() {
        return new HashMap<String, Long>(lagMonitoringMap);
    }

    /**
     * Set a list of static URLs to be replicated with, this is only necessary for nodes out of multicast range e.g. in
     * another data center.
     * 
     * @param urls A set of URLs, each one of the form ip:port. DNS addresses are also allowed.
     */
    public void setStaticUrls(Set<String> urls) {
        nodeUrlDao.setStaticUrls(urls);
    }

    /**
     * Set the port this server listens on for replication requests.
     * 
     * Default is to dynamically choose a random port.
     * 
     * @param port The port this server listens on for replication requests.
     */
    public void setPort(Integer port) {
        this.port = port;
    }

    /**
     * Set the amount of time between polling other nodes for map entry updates.
     * 
     * Default is 30 seconds.
     * 
     * @param entryPollingPeriod The entry polling period in ms.
     */
    public void setEntryPollingPeriod(Integer entryPollingPeriod) {
        this.entryPollingPeriod = entryPollingPeriod;
    }

    /**
     * Set the amount of time between polling for common maintenance tasks such as cleaning out dead nodes and
     * discovering other nodes.
     * 
     * Default is 60 seconds.
     * 
     * @param taskPollingPeriod The task polling period in ms.
     */
    public void setTaskPollingPeriod(Integer taskPollingPeriod) {
        this.taskPollingPeriod = taskPollingPeriod;
    }

    /**
     * Set the connection/read socket timeout when talking to other nodes.
     * 
     * Default is 10 seconds.
     * 
     * @param socketTimeout The socket timeout in ms.
     */
    public void setSocketTimeout(Integer socketTimeout) {
        this.socketTimeout = socketTimeout;
    }

    /**
     * Set how old deleted entries have to be before they are cleaned from the underlying database. Deleted entries are
     * kept to allow their deletion to be replicated to other nodes. If this value ever becomes less than the worst node
     * lag in the cluster then deletes will be lost.
     * 
     * Default is 1 week.
     * 
     * @param trimDeletedEntriesOlderThan How old deleted entries have to be before they are cleaned from the underlying
     *            database in ms.
     */
    public void setTrimDeletedEntriesOlderThan(Integer trimDeletedEntriesOlderThan) {
        this.trimDeletedEntriesOlderThan = trimDeletedEntriesOlderThan;
    }

    /**
     * Set how long a URL that was being advertised must go without an advertisement before it is removed from this
     * servers list of nodes to talk to.
     * 
     * Default is 5 minutes.
     * 
     * @param trimUnadvertisedUrlsOlderThan How long a URL that was being advertised must go without an advertisement
     *            before it is removed in ms.
     */
    public void setTrimUnadvertisedUrlsOlderThan(Integer trimUnadvertisedUrlsOlderThan) {
        this.trimUnadvertisedUrlsOlderThan = trimUnadvertisedUrlsOlderThan;
    }

    /**
     * Start this server.
     */
    public void start() {
        Collection<HttpResource> resources = new ArrayList<HttpResource>();
        resources.add(new EntryResource(entryDao));
        resources.add(new MapResource(this, entryDao));
        resources.add(new LagResource(this));
        httpServer = new HttpServer(context, node, port, resources);

        nodeUrlDiscovery = new NodeUrlDiscovery(context, node, nodeUrlDao, httpServer);

        entryPoller = new EntryPoller(context, node, nodeUrlDao, stateDao, entryDao, entryPollingPeriod, socketTimeout);

        periodicTasks = new PeriodicTask(context + "-" + getClass().getSimpleName(), taskPollingPeriod) {
            @Override
            public void doTask() {
                long trimTime = System.currentTimeMillis() - trimDeletedEntriesOlderThan;
                log.debug(String.format("Trimming deleted entries older than [%tc]", new Date(trimTime)));
                entryDao.trimOldEntries(trimTime);

                trimTime = System.currentTimeMillis() - trimUnadvertisedUrlsOlderThan;
                log.debug(String.format("Trimming unadvertised URLs older than [%tc]", new Date(trimTime)));
                nodeUrlDao.trimOldUrls(trimTime);

                if (Boolean.TRUE.equals(lagMonitoringMap.get(node))) {
                    log.debug("Node has written and entry at some point, writing a periodic entry so lag can be monitored");
                    lagMonitoringMap.put(node, System.currentTimeMillis());
                }

                nodeUrlDiscovery.ping();
            }
        };

        httpServer.start();
        nodeUrlDiscovery.start();
        entryPoller.start();
        periodicTasks.start();
        log.info(String.format("Started replicated map server for context [%s], node [%s]", context, node));
    }

    /**
     * Stop this server.
     */
    public void stop() {
        // TODO can we make the http server a daemon?
        httpServer.stop();
        log.info(String.format("Stopped replicated map server for context [%s], node [%s]", context, node));
    }

}
