package net.kierenb.mapcast.discovery;

import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.sql.SQLSyntaxErrorException;
import java.util.Set;

import net.kierenb.mapcast.sql.ResultTransformer;
import net.kierenb.mapcast.sql.SqlUtil;
import net.kierenb.mapcast.sql.StatementPreparator;

/**
 * Copyright 2012 Kieren Beckett
 * 
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * 
 * http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
 * an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
 * specific language governing permissions and limitations under the License.
 * 
 * @author mail@kierenb.net
 */
public class NodeUrlDao {

    private SqlUtil sqlUtil;

    public NodeUrlDao(SqlUtil sqlUtil) {
        this.sqlUtil = sqlUtil;

        try {
            sqlUtil.executeUpdate("CREATE TABLE url (url VARCHAR(255), lastAdvertised BIGINT, lastUsed BIGINT, PRIMARY KEY(url))");
        }
        catch (RuntimeException e) {
            // the table already exists
            if (!(e.getCause() instanceof SQLSyntaxErrorException)) {
                throw e;
            }
        }
    }

    public String getLeastRecentlyUsedUrl() {
        return sqlUtil.executeQuery("SELECT url FROM url ORDER BY lastUsed ASC LIMIT 1", new ResultTransformer<String>() {
            @Override
            public String transform(ResultSet rs) throws SQLException {
                if (rs.next()) {
                    return rs.getString("url");
                }
                return null;
            }
        });
    }

    public void setUsed(final String url) {
        sqlUtil.executeUpdate("UPDATE url SET lastUsed=? WHERE url=?", new StatementPreparator() {
            @Override
            public void prepare(PreparedStatement stmt) throws SQLException {
                int i = 1;
                stmt.setLong(i++, System.currentTimeMillis());
                stmt.setString(i++, url);
            }
        });
    }

    public void putUrl(final String url) {
        sqlUtil.executeUpdate("MERGE INTO url USING (VALUES(?, CAST(? AS BIGINT), CAST(-1 AS BIGINT))) AS new(url,lastAdvertised,lastUsed) ON url.url = new.url WHEN MATCHED THEN UPDATE SET url.lastAdvertised = new.lastAdvertised WHEN NOT MATCHED THEN INSERT VALUES new.url, new.lastAdvertised, new.lastUsed", new StatementPreparator() {
            @Override
            public void prepare(PreparedStatement stmt) throws SQLException {
                int i = 1;
                stmt.setString(i++, url);
                stmt.setLong(i++, System.currentTimeMillis());
            }
        });
    }

    public void setStaticUrls(Set<String> urls) {
        sqlUtil.executeUpdate("DELETE FROM url WHERE lastAdvertised=-1");

        for (final String url : urls) {
            sqlUtil.executeUpdate("MERGE INTO url USING (VALUES(?, CAST(-1 AS BIGINT), CAST(-1 AS BIGINT))) AS new(url,lastAdvertised,lastUsed) ON url.url = new.url WHEN MATCHED THEN UPDATE SET url.lastAdvertised = new.lastAdvertised WHEN NOT MATCHED THEN INSERT VALUES new.url, new.lastAdvertised, new.lastUsed", new StatementPreparator() {
                @Override
                public void prepare(PreparedStatement stmt) throws SQLException {
                    int i = 1;
                    stmt.setString(i++, url);
                }
            });
        }
    }

    public void trimOldUrls(final long trimTime) {
        sqlUtil.executeUpdate("DELETE FROM url WHERE lastAdvertised >= 0 AND lastAdvertised < ?", new StatementPreparator() {
            @Override
            public void prepare(PreparedStatement stmt) throws SQLException {
                int i = 1;
                stmt.setLong(i++, trimTime);
            }
        });
    }

}
