package net.kierenb.mapcast.discovery;

import java.io.IOException;
import java.net.DatagramPacket;
import java.net.InetAddress;
import java.net.InetSocketAddress;
import java.net.MulticastSocket;
import java.util.Arrays;

import net.kierenb.mapcast.util.HttpServer;

import org.apache.log4j.Logger;

/**
 * Copyright 2012 Kieren Beckett
 * 
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * 
 * http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
 * an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
 * specific language governing permissions and limitations under the License.
 * 
 * @author mail@kierenb.net
 */
public class NodeUrlDiscovery implements Runnable {
    private static final Logger log = Logger.getLogger(NodeUrlDiscovery.class);
    private static final String HOST = "239.54.113.10";
    private static final Integer PORT = 4446;
    private static final String COMMAND_PING = "PING";

    private String context;
    private String node;
    private NodeUrlDao nodeUrlDao;
    private HttpServer httpServer;
    private Thread thread;
    private MulticastSocket socket;
    private InetAddress group;

    public NodeUrlDiscovery(String context, String node, NodeUrlDao nodeUrlDao, HttpServer httpServer) {
        this.context = context;
        this.node = node;
        this.nodeUrlDao = nodeUrlDao;
        this.httpServer = httpServer;
    }

    public void start() {
        try {
            socket = new MulticastSocket(PORT);
            group = InetAddress.getByName(HOST);
            socket.joinGroup(group);
        }
        catch (IOException e) {
            throw new RuntimeException(e);
        }
        thread = new Thread(this, context + "-" + getClass().getSimpleName());
        thread.setDaemon(true);
        thread.start();
    }

    @Override
    public void run() {
        try {
            DatagramPacket packet;
            while (!thread.isInterrupted()) {
                byte[] buf = new byte[256];
                packet = new DatagramPacket(buf, buf.length);
                socket.receive(packet);
                String data = new String(Arrays.copyOf(packet.getData(), packet.getLength()), "UTF-8");
                InetSocketAddress address = (InetSocketAddress)packet.getSocketAddress();
                receive(address.getAddress().getHostAddress(), data);
            }
        }
        catch (IOException e) {
            throw new RuntimeException(e);
        }
        finally {
            if (socket != null) {
                try {
                    socket.leaveGroup(group);
                }
                catch (IOException e) {
                    log.error("Error shutting down, meh", e);
                }

                socket.close();
            }
        }
    }

    public void ping() {
        log.info("Sending ping");
        send(COMMAND_PING + " " + context + " " + node + " " + httpServer.getPort());
    }

    private void send(String data) {
        try {
            byte[] bytes = data.getBytes("UTF-8");
            DatagramPacket packet = new DatagramPacket(bytes, bytes.length, group, PORT);
            socket.send(packet);
        }
        catch (IOException e) {
            throw new RuntimeException(e);
        }
    }

    private void receive(String host, String data) {
        String[] parts = data.split(" ");
        String command = parts[0];
        String remoteContext = parts[1];
        String remoteNode = parts[2];

        if (!context.equals(remoteContext)) {
            log.debug(String.format("Ignoring packet from other context [%s]", remoteContext));
            return;
        }

        if (node.equals(remoteNode)) {
            log.debug(String.format("Ignoring packet from our node"));
            return;
        }

        if (COMMAND_PING.equals(command)) {
            Long port = Long.valueOf(parts[3]);
            String url = host + ":" + port;
            log.info(String.format("Got a PING command, storing URL [%s]", url));
            nodeUrlDao.putUrl(url);
        }
        else {
            log.warn(String.format("Unrecognised packet data [%s] from host [%s]", data, host));
        }
    }

}
