package net.kierenb.mapcast.replication;

import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.sql.SQLIntegrityConstraintViolationException;
import java.sql.SQLSyntaxErrorException;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;

import net.kierenb.mapcast.sql.ResultTransformer;
import net.kierenb.mapcast.sql.SqlUtil;
import net.kierenb.mapcast.sql.StatementPreparator;

/**
 * Copyright 2012 Kieren Beckett
 * 
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * 
 * http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
 * an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
 * specific language governing permissions and limitations under the License.
 * 
 * @author mail@kierenb.net
 */
public class EntryDao {
    public static final Integer MAX_NODE_LENGTH = 40;
    public static final Integer MAX_MAP_NAME_LENGTH = 20;
    public static final Integer MAX_KEY_LENGTH = 40;

    private SqlUtil sqlUtil;

    public EntryDao(SqlUtil sqlUtil) {
        this.sqlUtil = sqlUtil;

        try {
            sqlUtil.executeUpdate("CREATE TABLE entry (map VARCHAR(" + MAX_MAP_NAME_LENGTH + "), key VARCHAR(" + MAX_KEY_LENGTH + "), value BLOB, deletedTime BIGINT, lastUpdatedNode VARCHAR(" + MAX_NODE_LENGTH + "), lastUpdatedTime BIGINT, PRIMARY KEY(map, key))");
            sqlUtil.executeUpdate("CREATE INDEX update ON entry (lastUpdatedNode, lastUpdatedTime)");
            sqlUtil.executeUpdate("CREATE INDEX trim ON entry (deletedTime)");
        }
        catch (RuntimeException e) {
            // the table already exists
            if (!(e.getCause() instanceof SQLSyntaxErrorException)) {
                throw e;
            }
        }
    }

    public Collection<ReplicatedMapEntry> getEntriesSince(final String excludeNode, final Map<String, Long> sinceTimes, final Integer limit) {
        String criteria = "lastUpdatedNode NOT IN (?";
        for (int i = 0; i < sinceTimes.size(); i++) {
            criteria += ",?";
        }
        criteria += ")";
        for (int i = 0; i < sinceTimes.size(); i++) {
            criteria += " OR (lastUpdatedNode=? AND lastUpdatedTime > ?)";
        }

        return sqlUtil.executeQuery("SELECT map, key, value, deletedTime, lastUpdatedNode, lastUpdatedTime FROM entry WHERE " + criteria + " ORDER BY lastUpdatedTime ASC LIMIT ?", new StatementPreparator() {
            @Override
            public void prepare(PreparedStatement stmt) throws SQLException {
                int i = 1;
                stmt.setString(i++, excludeNode);
                for (String node : sinceTimes.keySet()) {
                    stmt.setString(i++, node);
                }
                for (Map.Entry<String, Long> sinceTime : sinceTimes.entrySet()) {
                    stmt.setString(i++, sinceTime.getKey());
                    stmt.setLong(i++, sinceTime.getValue());
                }
                stmt.setInt(i++, limit);
            }
        }, ENTRY_TRANSFORMER);
    }

    public void putEntry(final ReplicatedMapEntry entry) {
        // We replace the existing entry if ours is newer or ours has exactly the same time but a greater node ID. The
        // last bit is simply to make updating deterministic.

        try {
            sqlUtil.executeUpdate("MERGE INTO entry USING (VALUES(?, ?, CAST(? AS BLOB), CAST(? AS BIGINT), ?, CAST(? AS BIGINT))) AS new(map,key,value,deletedTime,lastUpdatedNode,lastUpdatedTime) ON entry.map = new.map AND entry.key = new.key AND (entry.lastUpdatedTime < new.lastUpdatedTime OR entry.lastUpdatedTime = new.lastUpdatedTime AND entry.lastUpdatedNode < new.lastUpdatedNode) WHEN MATCHED THEN UPDATE SET entry.value = new.value, entry.deletedTime=new.deletedTime, entry.lastUpdatedNode = new.lastUpdatedNode, entry.lastUpdatedTime = new.lastUpdatedTime WHEN NOT MATCHED THEN INSERT VALUES new.map, new.key, new.value, new.deletedTime, new.lastUpdatedNode, new.lastUpdatedTime", new StatementPreparator() {
                @Override
                public void prepare(PreparedStatement stmt) throws SQLException {
                    int i = 1;
                    stmt.setString(i++, entry.getMap());
                    stmt.setString(i++, entry.getKey());
                    stmt.setBytes(i++, entry.getValue());
                    stmt.setLong(i++, entry.getDeletedTime());
                    stmt.setString(i++, entry.getLastUpdatedNode());
                    stmt.setLong(i++, entry.getLastUpdatedTime());
                }
            });
        }
        catch (RuntimeException e) {
            // In the above we update where the key matches and the entry time is newer than what exists.
            // If the key doesn't match a row or the entry time is newer we try to insert, this is fine when the key
            // doesn't exist.
            // When the key does exist unfortunately we get a duplicate key exception, this means there's a newer entry
            // already and we should silently ignore the exception.
            if (!(e.getCause() instanceof SQLIntegrityConstraintViolationException)) {
                throw e;
            }
        }
    }

    public void trimOldEntries(final Long trimTime) {
        sqlUtil.executeUpdate("DELETE FROM entry WHERE deletedTime >= 0 AND deletedTime < ?", new StatementPreparator() {
            @Override
            public void prepare(PreparedStatement stmt) throws SQLException {
                int i = 1;
                stmt.setLong(i++, trimTime);
            }
        });
    }
    
    public Set<String> getMaps() {
        return sqlUtil.executeQuery("SELECT DISTINCT map FROM entry", new ResultTransformer<Set<String>>() {
            @Override
            public Set<String> transform(ResultSet rs) throws SQLException {
                Set<String> maps = new HashSet<String>();
                while (rs.next()) {
                    maps.add(rs.getString("map"));
                }
                return maps;
            }
        });
    }

    private static final ResultTransformer<Collection<ReplicatedMapEntry>> ENTRY_TRANSFORMER = new ResultTransformer<Collection<ReplicatedMapEntry>>() {
        @Override
        public Collection<ReplicatedMapEntry> transform(ResultSet rs) throws SQLException {
            Collection<ReplicatedMapEntry> entries = new ArrayList<ReplicatedMapEntry>();
            while (rs.next()) {
                String map = rs.getString("map");
                String key = rs.getString("key");
                byte[] value = rs.getBytes("value");
                Long deletedTime = rs.getLong("deletedTime");
                String lastUpdatedNode = rs.getString("lastUpdatedNode");
                Long lastUpdatedTime = rs.getLong("lastUpdatedTime");
                entries.add(new ReplicatedMapEntry(map, key, value, deletedTime, lastUpdatedNode, lastUpdatedTime));
            }
            return entries;
        }
    };

}
