package net.kierenb.mapcast.replication;

import java.io.IOException;
import java.io.InputStream;
import java.io.UnsupportedEncodingException;
import java.net.URL;
import java.net.URLConnection;
import java.net.URLEncoder;
import java.util.Map;

import javax.mail.BodyPart;
import javax.mail.MessagingException;
import javax.mail.Multipart;
import javax.mail.internet.MimeMultipart;
import javax.mail.util.ByteArrayDataSource;

import net.kierenb.mapcast.discovery.NodeUrlDao;
import net.kierenb.mapcast.util.HttpServer;

import org.apache.log4j.Logger;
import org.apache.log4j.lf5.util.StreamUtils;

/**
 * Copyright 2012 Kieren Beckett
 * 
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * 
 * http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
 * an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
 * specific language governing permissions and limitations under the License.
 * 
 * @author mail@kierenb.net
 */
public class EntryPoller implements Runnable {
    private static final Logger log = Logger.getLogger(EntryPoller.class);

    private String context;
    private String node;
    private NodeUrlDao nodeUrlDao;
    private StateDao stateDao;
    private EntryDao entryDao;
    private Integer pollingPeriod;
    private Integer socketTimeout;
    private Thread thread;

    public EntryPoller(String context, String node, NodeUrlDao nodeUrlDao, StateDao stateDao, EntryDao entryDao, Integer pollingPeriod, Integer socketTimeout) {
        this.context = context;
        this.node = node;
        this.nodeUrlDao = nodeUrlDao;
        this.stateDao = stateDao;
        this.entryDao = entryDao;
        this.pollingPeriod = pollingPeriod;
        this.socketTimeout = socketTimeout;
    }

    public void start() {
        thread = new Thread(this, context + "-" + getClass().getSimpleName());
        thread.setDaemon(true);
        thread.start();
    }

    @Override
    public void run() {
        while (!thread.isInterrupted()) {
            String baseUrl = nodeUrlDao.getLeastRecentlyUsedUrl();
            if (baseUrl == null) {
                log.info("Couldn't find any URLs to poll");
            }
            else {
                boolean moreToProcess = true;
                while (!thread.isInterrupted() && moreToProcess) {
                    log.info(String.format("Polling %s", baseUrl));
                    moreToProcess = process(baseUrl);
                }
                // TODO if request failed should we avoid using that url again for a bit?
                nodeUrlDao.setUsed(baseUrl);
            }

            try {
                log.debug("Pausing before polling again");
                Thread.sleep(pollingPeriod);
            }
            catch (InterruptedException e) {
                thread.interrupt();
            }
        }
    }

    private boolean process(String baseUrl) {
        Map<String, Long> state = stateDao.getState();
        String url = baseUrl + EntryResource.URL + "?excludeNode=" + urlEncode(node) + getSinceTimes(state);
        InputStream in = null;
        try {
            URLConnection conn = new URL("http://" + url).openConnection();
            // TODO do the timeouts work?
            conn.setConnectTimeout(socketTimeout);
            conn.setReadTimeout(socketTimeout);
            in = conn.getInputStream();
            String remoteContext = conn.getHeaderField(HttpServer.HEADER_CONTEXT);
            String remoteNode = conn.getHeaderField(HttpServer.HEADER_NODE);

            if (!context.equals(remoteContext)) {
                log.error(String.format("URL [%s] is for other context [%s], this shouldn't happen", baseUrl, remoteContext));
                return false;
            }

            if (node.equals(remoteNode)) {
                log.error(String.format("URL [%s] is for our node, this shouldn't happen", baseUrl));
                return false;
            }

            return process(in, conn.getContentType());
        }
        catch (Exception e) {
            log.warn(String.format("Error making request to [http://%s]", url), e);
            return false;
        }
        finally {
            if (in != null) {
                try {
                    in.close();
                }
                catch (IOException e) {
                    log.error("Error closing connection, meh", e);
                }
            }
        }
    }

    private Boolean process(InputStream in, String contentType) throws IOException {
        try {
            if (contentType == null) {
                // no entries to fetch
                return false;
            }

            Multipart multipart = new MimeMultipart(new ByteArrayDataSource(in, contentType));
            for (int i = 0; i < multipart.getCount(); i++) {
                BodyPart part = multipart.getBodyPart(i);
                String map = part.getHeader(EntryResource.HEADER_MAP)[0];
                String key = part.getHeader(EntryResource.HEADER_KEY)[0];
                Long deletedTime = Long.valueOf(part.getHeader(EntryResource.HEADER_DELETED)[0]);
                String lastUpdatedNode = part.getHeader(EntryResource.HEADER_LAST_UPDATED_NODE)[0];
                Long lastUpdatedTime = Long.valueOf(part.getHeader(EntryResource.HEADER_LAST_UPDATED_TIME)[0]);
                byte[] value = StreamUtils.getBytes(part.getInputStream());

                ReplicatedMapEntry entry = new ReplicatedMapEntry(map, key, value, deletedTime, lastUpdatedNode, lastUpdatedTime);
                entryDao.putEntry(entry);
                stateDao.updateState(lastUpdatedNode, lastUpdatedTime);
            }

            return multipart.getCount() >= EntryResource.MAX_ENTRIES_PER_POLL;
        }
        catch (MessagingException e) {
            log.error("Error parsing mime entries", e);
            return false;
        }
    }

    private static String getSinceTimes(Map<String, Long> state) {
        if (state.size() == 0) {
            return "";
        }

        String sinceTimes = "";
        for (Map.Entry<String, Long> entry : state.entrySet()) {
            sinceTimes += entry.getKey() + ":" + entry.getValue() + ",";
        }
        sinceTimes = sinceTimes.substring(sinceTimes.length() - 1);
        return "&since=" + urlEncode(sinceTimes);
    }

    private static String urlEncode(String str) {
        try {
            return URLEncoder.encode(str, "UTF-8");
        }
        catch (UnsupportedEncodingException e) {
            // not going to happen
            throw new RuntimeException(e);
        }
    }

}
