package net.kierenb.mapcast.replication;

import java.io.IOException;
import java.io.OutputStream;
import java.util.HashMap;
import java.util.Map;

import javax.activation.DataHandler;
import javax.mail.MessagingException;
import javax.mail.internet.MimeBodyPart;
import javax.mail.internet.MimeMultipart;
import javax.mail.util.ByteArrayDataSource;

import net.kierenb.mapcast.util.HttpResource;

import org.apache.log4j.Logger;
import org.simpleframework.http.Request;
import org.simpleframework.http.Response;

/**
 * Copyright 2012 Kieren Beckett
 * 
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * 
 * http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
 * an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
 * specific language governing permissions and limitations under the License.
 * 
 * @author mail@kierenb.net
 */
public class EntryResource implements HttpResource {
    private static final Logger log = Logger.getLogger(EntryResource.class);

    public static final String URL = "/entries/";
    public static final String HEADER_MAP = "map";
    public static final String HEADER_KEY = "key";
    public static final String HEADER_DELETED = "deleted";
    public static final String HEADER_LAST_UPDATED_NODE = "lastUpdatedNode";
    public static final String HEADER_LAST_UPDATED_TIME = "lastUpdatedTime";
    public static final Integer MAX_ENTRIES_PER_POLL = 50;

    private EntryDao entryDao;

    public EntryResource(EntryDao entryDao) {
        this.entryDao = entryDao;
    }

    @Override
    public String getUrl() {
        return URL;
    }

    @Override
    public void handle(Request req, Response resp) throws IOException {
        if (!"GET".equals(req.getMethod())) {
            resp.setCode(405);
            return;
        }
        
        if (!URL.equals(req.getPath().getPath())) {
            resp.setCode(404);
            return;
        }

        try {
            Map<String, Long> sinceTimes = getSinceTimes(req.getParameter("since"));
            String excludeNode = req.getParameter("excludeNode");
            OutputStream out = resp.getOutputStream();
            MimeMultipart multipart = new MimeMultipart();
            for (ReplicatedMapEntry entry : entryDao.getEntriesSince(excludeNode, sinceTimes, MAX_ENTRIES_PER_POLL)) {
                MimeBodyPart part = new MimeBodyPart();
                part.setHeader(HEADER_MAP, entry.getMap());
                part.setHeader(HEADER_KEY, entry.getKey());
                part.setHeader(HEADER_DELETED, entry.getDeletedTime().toString());
                part.setHeader(HEADER_LAST_UPDATED_NODE, entry.getLastUpdatedNode());
                part.setHeader(HEADER_LAST_UPDATED_TIME, entry.getLastUpdatedTime().toString());
                part.setDataHandler(new DataHandler(new ByteArrayDataSource(entry.getValue(), "application/octet-stream")));
                multipart.addBodyPart(part);
            }

            if (multipart.getCount() > 0) {
                resp.set("content-type", multipart.getContentType());
                multipart.writeTo(out);
            }
            else {
                resp.setCode(201);
            }
        }
        catch (MessagingException e) {
            log.error(String.format("Error generating mime for request [%s]", req.getPath().getPath()), e);
        }
    }

    private static Map<String, Long> getSinceTimes(String sinceTimesStr) {
        Map<String, Long> sinceTimes = new HashMap<String, Long>();
        if (sinceTimesStr != null && sinceTimesStr.length() > 0) {
            for (String sinceTimeStr : sinceTimesStr.split(",")) {
                String[] nodeAndTime = sinceTimeStr.split(":");
                String node = nodeAndTime[0];
                Long time = Long.valueOf(nodeAndTime[1]);
                sinceTimes.put(node, time);
            }
        }
        return sinceTimes;
    }

}
