package net.kierenb.mapcast.sql;

import java.sql.Connection;
import java.sql.DriverManager;
import java.sql.PreparedStatement;
import java.sql.SQLException;

import org.apache.log4j.Logger;

/**
 * Copyright 2012 Kieren Beckett
 * 
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * 
 * http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
 * an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
 * specific language governing permissions and limitations under the License.
 * 
 * @author mail@kierenb.net
 */
public class SqlUtil {
    private static final Logger log = Logger.getLogger(SqlUtil.class);

    private String filename;

    public SqlUtil(String filename) {
        this.filename = filename;

        try {
            Class.forName("org.hsqldb.jdbc.JDBCDriver");
        }
        catch (ClassNotFoundException e) {
            throw new RuntimeException("Cannot load HSQLDB driver");
        }
    }

    public Integer executeUpdate(String sql) {
        return executeUpdate(sql, NULL_PREPARATOR);
    }

    public Integer executeUpdate(String sql, StatementPreparator preparator) {
        return executeSql(sql, preparator, UPDATE_EXECUTOR);
    }

    public <R> R executeQuery(String sql, ResultTransformer<R> transformer) {
        return executeQuery(sql, NULL_PREPARATOR, transformer);
    }

    public <R> R executeQuery(String sql, StatementPreparator preparator, ResultTransformer<R> transformer) {
        return executeSql(sql, preparator, new QueryExecutor<R>(transformer));
    }

    private <R> R executeSql(String sql, StatementPreparator preparator, StatementExecutor<R> transformer) {
        long startTime = System.currentTimeMillis();
        Connection conn = null;
        try {
            conn = DriverManager.getConnection("jdbc:hsqldb:file:" + filename, "SA", "");
            PreparedStatement stmt = conn.prepareStatement(sql);
            preparator.prepare(stmt);
            return transformer.execute(stmt);
        }
        catch (SQLException e) {
            throw new RuntimeException(e);
        }
        finally {
            if (conn != null) {
                try {
                    conn.close();
                }
                catch (SQLException e) {
                    throw new RuntimeException(e);
                }
            }
            log.trace(String.format("Executed SQL [%s] in %d ms", sql, System.currentTimeMillis() - startTime));
        }
    }

    private static interface StatementExecutor<R> {
        public R execute(PreparedStatement rs) throws SQLException;
    }

    private static class QueryExecutor<R> implements StatementExecutor<R> {
        private ResultTransformer<R> transformer;

        public QueryExecutor(ResultTransformer<R> transformer) {
            this.transformer = transformer;
        }

        @Override
        public R execute(PreparedStatement stmt) throws SQLException {
            return transformer.transform(stmt.executeQuery());
        }
    };

    private static final StatementPreparator NULL_PREPARATOR = new StatementPreparator() {
        @Override
        public void prepare(PreparedStatement stmt) throws SQLException {

        }
    };

    private static final StatementExecutor<Integer> UPDATE_EXECUTOR = new StatementExecutor<Integer>() {
        @Override
        public Integer execute(PreparedStatement stmt) throws SQLException {
            return stmt.executeUpdate();
        }
    };

}
