package net.kierenb.mapcast.util;

import java.io.IOException;
import java.net.InetSocketAddress;
import java.util.Collection;

import org.apache.log4j.Logger;
import org.simpleframework.http.Request;
import org.simpleframework.http.Response;
import org.simpleframework.http.core.Container;
import org.simpleframework.transport.connect.Connection;
import org.simpleframework.transport.connect.SocketConnection;

/**
 * Copyright 2012 Kieren Beckett
 * 
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * 
 * http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
 * an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
 * specific language governing permissions and limitations under the License.
 * 
 * @author mail@kierenb.net
 */
public class HttpServer implements Container {
    private static final Logger log = Logger.getLogger(HttpServer.class);

    public static final String HEADER_CONTEXT = "X-mapcast-context";
    public static final String HEADER_NODE = "X-mapcast-node";

    private String context;
    private String node;
    private Collection<HttpResource> resources;
    private Connection connection;
    private Integer port;

    public HttpServer(String context, String node, Integer port, Collection<HttpResource> resources) {
        this.context = context;
        this.node = node;
        this.port = port;
        this.resources = resources;
    }

    public Integer getPort() {
        return port;
    }

    public void start() {
        try {
            connection = new SocketConnection(this);
            InetSocketAddress address = new InetSocketAddress(port);
            address = (InetSocketAddress)connection.connect(address);
            port = address.getPort();
            log.info(String.format("Started replicated map HTTP server on port [%d]", port));
        }
        catch (IOException e) {
            throw new RuntimeException(e);
        }
    }

    public void stop() {
        try {
            connection.close();
        }
        catch (IOException e) {
            log.error("Error shutting down, meh", e);
        }
    }

    @Override
    public void handle(Request req, Response resp) {
        long startTime = System.currentTimeMillis();
        String path = req.getPath().getPath();
        String ip = req.getClientAddress().toString();
        try {
            resp.add(HEADER_CONTEXT, context);
            resp.add(HEADER_NODE, node);
            HttpResource resource = getResource(path);
            if (resource != null) {
                resource.handle(req, resp);
            }
            else {
                log.warn(String.format("Unrecognised request [%s] for IP [%s], returning 404", path, ip));
                resp.setCode(404);
            }
        }
        catch (Exception e) {
            log.error(String.format("Error handling request [%s] for IP [%s]", path, ip), e);
        }
        finally {
            try {
                resp.getOutputStream().close();
            }
            catch (IOException e) {
                log.error("Error closing connection, meh", e);
            }
            log.debug(String.format("Handled request [%s] for IP [%s] in %d ms", path, ip, System.currentTimeMillis() - startTime));
        }
    }

    private HttpResource getResource(String path) {
        for (HttpResource resource : resources) {
            if (path.startsWith(resource.getUrl())) {
                return resource;
            }
        }
        return null;
    }

}
